#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <unistd.h>

#include <libusb-1.0/libusb.h>

static libusb_context *context = NULL;
static libusb_device_handle *handle = NULL;
static libusb_device_handle *handles[8];

#define FRAMESIZE5 (854 * 480 * 2)
#define FRAMESIZE  (800 * 480 * 2)
#define MARGIN5    (320)
#define MARGIN     (0)


int get_version(libusb_device_handle *h)
{
    unsigned char buf[5] = {0x51, 0x02, 0x04, 0x1f, 0xfc};
    libusb_control_transfer(h, 0x40, 0xb5, 0, 0, buf, 5, 100);
    libusb_control_transfer(h, 0xc0, 0xb6, 0, 0, buf, 1, 100);
    libusb_control_transfer(h, 0xc0, 0xb7, 0, 0, buf, 5, 100);
    return ((int *)(buf + 1))[0];
}

void get_screen_version(libusb_device_handle *h, int *frame_size, int *margin, char *buf)
{
    switch (get_version(h)) {
    case 0x00000005:
        if (buf)
            sprintf(buf, "5inch, 480x854x16/24, SLM5.0-81FPC-A");
        if (frame_size)
            *frame_size = FRAMESIZE5;
        if (margin)
            *margin = MARGIN5;
        break;

    case 0x00000304:
        if (buf)
            sprintf(buf, "4.3inch, 480x800x16/24, D430FPC9316-A");
        if (frame_size)
            *frame_size = FRAMESIZE;
        if (margin)
            *margin = MARGIN;
        break;

    case 0x00000004:
        if (buf)
            sprintf(buf, "4inch, 480x800x16, TOSHIBA-2122");
        if (frame_size)
            *frame_size = FRAMESIZE;
        if (margin)
            *margin = MARGIN;
        break;

    case 0x00000104:
        if (buf)
            sprintf(buf, "4inch, 480x800x16/24, DJN-1922");
        if (frame_size)
            *frame_size = FRAMESIZE;
        if (margin)
            *margin = MARGIN;
        break;

    default:
        if (buf)
            sprintf(buf, "4inch/5inch, 480x800x16, default type");
        if (frame_size)
            *frame_size = FRAMESIZE;
        if (margin)
            *margin = MARGIN;
        break;
    }
}

void get_screen_id(libusb_device_handle *h, unsigned char *id)
{
    unsigned char cmd[] = {0x51, 0x02, 0x08, 0x1f, 0xf0};
    unsigned char buf[9] = {0};
    
    libusb_control_transfer(h, 0x40, 0xb5, 0, 0, cmd, 5, 100);
    libusb_control_transfer(h, 0xc0, 0xb6, 0, 0, cmd, 1, 100);
    libusb_control_transfer(h, 0xc0, 0xb7, 0, 0, buf, 9, 100);
    memcpy(id, buf + 1, 8);
}

void set_screen_id(libusb_device_handle *h, const unsigned char *id)
{
    unsigned char cmd[13] = {0x51, 0x0a, 0x00, 0x1f, 0xf0};
    unsigned char buf[1] = {0xef}; // enable write flash
    
    memcpy(cmd + 5, id, 8);
    libusb_control_transfer(h, 0x40, 0xb9, 0, 0, buf, 1, 100);    
    libusb_control_transfer(h, 0x40, 0xb5, 0, 0, cmd, 13, 100);
    libusb_control_transfer(h, 0xc0, 0xb6, 0, 0, cmd, 1, 100);
}

void print_screen_id(libusb_device_handle *h)
{
    unsigned char id[8] = {0};
    get_screen_id(h, id);
    printf("id: %02x%02x-%02x%02x-%02x%02x%02x%02x\n", 
        id[0], id[1], id[2], id[3], id[4], id[5], id[6], id[7]);
}

int get_port_handle(uint16_t vid, uint16_t pid)
{
    struct libusb_device **list = NULL;
    int used = 0;

    int count = libusb_get_device_list(context, &list);
    for (int i = 0; i < count || used > 8; i++) {
        struct libusb_device_descriptor desc;

        memset(&desc, 0, sizeof(struct libusb_device_descriptor));
        if (libusb_get_device_descriptor(list[i], &desc))
            continue;

        int curport = libusb_get_port_number(list[i]) - 1;
        if (desc.idVendor == vid && desc.idProduct == pid) {
            char screen_info[0x100] = {0};

            libusb_open(list[i], &handles[used]);

            libusb_claim_interface(handles[used], 0);
            libusb_set_interface_alt_setting(handles[used], 0, 0);

            get_screen_version(handles[used], NULL, NULL, screen_info);

            printf("Find screen at USB physics port %d\n", curport);
            printf("%d: Type: %s\n", used + 1, screen_info);
            print_screen_id(handles[used]);
            
            used++;
        }
    }

    libusb_free_device_list(list, 0);
    return used;
}

int main(int argc, char *argv[])
{    
    libusb_init(&context);
    
    int count = get_port_handle(0xc872, 0x1004);
    if (count <= 0) {
        printf("error: no screen find.\n");
        return -1;
    }
    
    printf("\nchoose a screen by index to set random id(1~%d, other cancel): ", count);
    int id = getchar() - '1';
    if (id < 0 || id >= 8) {
        printf("error: invalid index.\n");
        return -1;
    }
    handle = handles[id];
    libusb_claim_interface(handle, 0);
    libusb_set_interface_alt_setting(handle, 0, 0);
    
    // generate random id.
    unsigned char buf[8];
    srand(time(NULL));
    buf[0] = 0xef;
    buf[1] = 0x00;
    buf[2] = rand();
    buf[3] = 0x00;
    buf[4] = rand() % 0xff;
    buf[5] = rand() % 0xff;
    buf[6] = rand() % 0xff;
    buf[7] = rand() % 0xff;
    
    // write the random id to flash.
    set_screen_id(handle, buf);
    
    printf("new ");
    print_screen_id(handle);
    printf("\n\n");
    
    libusb_release_interface(handle, 0);
    libusb_close(handle);
    libusb_exit(context);
    
    return 0;
}
